#define _USE_MATH_DEFINES

#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <algorithm>
#include <vector>

using namespace std;

template<class T> inline T sqr (T x) { return x * x; }

#define eps 1e-5
#define sz(C) (int)((C).size())

struct vect {
	float x, y;
	vect (float x = 0, float y = 0)
		: x (x), y (y) {}
	float md () const {
		return _hypot(x, y);
	}
	vect operator - (const vect &b) const {
		return vect(x - b.x, y - b.y);
	}
};

struct point {
	float x, y;
	point (float x = 0, float y = 0)
		: x (x), y (y) {}
};

const int SZ = 32;
const float R = (sqrt(5.0) - 1.0) / 4.0;
const float min_dist = 0.08;

World *world;
int ms[1 << 17], mn = 0;
point player;

inline float p2p (point a, point b) {
	return _hypot(a.x - b.x, a.y - b.y);
}

inline float dist (float x1, float y1, float x2, float y2) {
	return _hypot(x1 - x2, y1 - y2);
}

float closest;
vect dir;
bool has_bullets;

inline bool bullets () {
	for (int i = 0; i < mn; ++i) {
		int j = ms[i];
		if (world->monsters[j].BulletDamage)
			return true;
	}
	return false;
}

inline int choose_monster () {
	int res = -1;
	float best_dist = 1e10, best_hp;
	for (int j = 0; j < mn; ++j) {
		int i = ms[j];
		float d = dist(world->monsters[i].PositionX, world->monsters[i].PositionY, world->player.PositionX, world->player.PositionY);
		float hp = world->monsters[i].HitPoints;
		if (res == -1 || best_dist > d || abs(best_dist - d) < eps && best_hp > hp) {
			res = i;
			best_dist = d;
			best_hp = hp;
		}
	}
	closest = best_dist;
	return res;
}

#define ITER 40

inline point get_point (point p, vect v, float t) {
	return point(p.x + v.x * t, p.y + v.y * t);
}

inline bool ok (point p) {
	return p.x >= 0.0 && p.y >= 0.0 && p.x <= 1.0 && p.y <= 1.0;
}

inline bool bsearch (point p, vect v, float &time) {
	float l = 0.0, r = 10000.0, m;
	for (int it = 0; it < ITER; ++it) {
		m = (l + r) / 2.0;
		point cur = get_point(p, v, m);
		float d = p2p(player, cur);
		time = d / world->player.BulletVelocity;
		if (time < m)
			r = m;
		else
			l = m;
	}
	point cur = get_point(p, v, time);
	return ok(cur);
}

void shoot (int has = -1) {
	int monster_id = has == -1 ? choose_monster() : has;
	if (monster_id == -1) return;
	float x = world->monsters[monster_id].PositionX - world->player.PositionX;
	float y = world->monsters[monster_id].PositionY - world->player.PositionY;
	world->shoot(x, y);
}

void smart_shoot () {
	if (!mn) return;
	int j = choose_monster();
	if (j == -1) return;
	float time;
	point p (world->monsters[j].PositionX, world->monsters[j].PositionY);
	vect v (world->monsters[j].VelocityX, world->monsters[j].VelocityY);
	if (!bsearch(p, v, time)) {
		shoot(j);
		return;
	}
	point cur = get_point(p, v, time);
	world->shoot(cur.x - player.x, cur.y - player.y);
}

void get_monsters () {
	mn = 0;
	for (int i = 0; i < world->MAXMONSTERS; ++i) {
		if (world->monsters[i].ClassName == -1) continue;
		ms[mn++] = i;
		if (mn == (1 << 17)) break;
	}
}

int cur = 0;
int sz_target = 8;
float tx[128] = {0.50, 0.90, 0.85, 0.80, 0.50, 0.20, 0.15, 0.10};
float ty[128] = {0.50, 0.50, 0.80, 0.85, 0.90, 0.85, 0.80, 0.50};

void next_target () {
	++cur %= sz_target;
}

bool phase2 = false;

void move () {
	if (!has_bullets) {
		static bool need2run = false;
		if (closest > min_dist && !need2run) return;
		need2run = true;
		float d = dist(tx[cur], ty[cur], player.x, player.y);
		if (fabs(d) < 1e-2) {
			next_target();
			if (cur == 0) need2run = false;
		}
		dir = vect(tx[cur] - player.x, ty[cur] - player.y);
		world->run(dir.x, dir.y);
	} else {
		if (!phase2) {
			sz_target = 3;
			cur = 0;
			tx[1] = 0.5, ty[1] = 0.1;
			tx[2] = 0.5, ty[2] = 0.9;
			phase2 = true;
		}
		float d = dist(tx[cur], ty[cur], player.x, player.y);
		if (fabs(d) < 1e-2) {
			next_target();
		}
		dir = vect(tx[cur] - player.x, ty[cur] - player.y);
		world->run(dir.x, dir.y);
	}
}

int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	world = new World(argv[1], atoi(argv[2]));
	

	//Infinite loop
	for (int i = 0; ; ++i) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

		player = point(world->player.PositionX, world->player.PositionY);
		has_bullets = bullets();
		get_monsters();
		smart_shoot();
		move();
	}

	return 0;
}

