#include "AI.h"
#include "Entities.h"
#include "GameWorld.h"

RandomAI::RandomAI(Monster *_owner, double _awarenessradius)
	 : AIBase(_owner, _awarenessradius), cooldown(0.0) {
}

void RandomAI::PassTime(double deltatime) {
	cooldown.PassTime(deltatime);
}

void RandomAI::Think(EntitySet &awareof) {
	if (cooldown.TimeToAct()) {
		cooldown.period = trandom()*5.0;
		double ang = trandom() * 4.0*acos(0.0);
		dir = double2(cos(ang), sin(ang));
	}
	owner->SetMoveDirection(dir);
}

Player *FindEnemy(AIBase::EntitySet &awareof) {
	for (int i = 0; i<awareof.size(); i++) {
		Entity *ent = awareof[i];
		if (IsPlayer(ent)) return (Player*)ent;
	}
	return 0;
}

double2 GetAntiCollisionForce(Monster *owner, AIBase::EntitySet &awareof) {
	double2 force;
	for (int i = 0; i<awareof.size(); i++) {
		Entity *ent = awareof[i];
		if (!IsMonster(ent)) continue;
		double2 toent = ent->position - owner->position;
		double normdist = !toent / (ent->radius + owner->radius);
		force -= ForceVectorNorm(toent, 1.0 / normdist);
	}
	return force;
}

double2 GetMissileDodgingForce(Monster *owner, AIBase::EntitySet &awareof) {
	double2 force;
	for (int i = 0; i<awareof.size(); i++) {
		Entity *ent = awareof[i];
		if (!IsMissile(ent)) continue;
		if ((ent->velocity ^ (owner->position - ent->position)) < 0) continue;

		double2 norm = ForceVectorNorm(ent->velocity, 1);
		norm = double2(-norm.y, norm.x);

		double hgt = (owner->position - ent->position) ^ norm;
		double level = ent->radius + owner->radius*3.0;
		if (fabs(hgt) > level) continue;

		force += norm * (hgt < 0 ? -1 : 1);
	}
	return force;
}

double2 GetAttractionForce(Monster *owner, Entity *ent) {
	if (ent) return ForceVectorNorm(ent->position - owner->position, 1.0);
	else return double2();
}

double2 GetInFieldForce(Monster *owner) {
	double2 p = owner->position / FIELD_SIZE;
	double bdst = std::min(std::min(p.x, p.y), std::min(1.0 - p.x, 1.0 - p.y));
	double rad = 0.1;
	double mult = exp(1.0 - (bdst/rad));
	return ForceVectorNorm(double2(0.5, 0.5) - p, mult);
}

void ShootGood(Monster *owner, const Player *pl) {
	if (owner->gun.get()) {
		double speed = owner->gun->bulletvelocity;
		double2 dir = ShootMovingTarget(pl->position - owner->position, pl->velocity * (rand()%3)*0.5, speed);
		owner->Shoot(dir);
	}
}



DroneAI::DroneAI(Monster *_owner, double _awarenessradius
	, double _thinkperiod, double _playerimportance, double _randomimportance)
	 : AIBase(_owner, _awarenessradius)
	 , cooldown(_thinkperiod)
	 , playerimportance(_playerimportance), randomimportance(_randomimportance) {
}

void DroneAI::PassTime(double deltatime) {
	cooldown.PassTime(deltatime);
}

void DroneAI::Think(EntitySet &awareof) {
	Player *pl = FindEnemy(awareof);
	if (cooldown.TimeToAct()) {
		double2 force;
		force += GetAttractionForce(owner, pl) * playerimportance;
		force += RandomVector() * randomimportance;
		force += GetAntiCollisionForce(owner, awareof);
		owner->SetMoveDirection(force);
	}
	ShootGood(owner, pl);
}


WallAI::WallAI(Monster *_owner, double _awarenessradius, double2 _direction)
	 : AIBase(_owner, _awarenessradius), direction(_direction) {
	desiredposition = owner->position;
	double fx = fabs(direction.x);
	double fy = fabs(direction.y);
	if (fx > fy) side = double2((direction.x > 0 ? 1 : -1), 0.0);
	else side = double2(0.0, (direction.y > 0 ? 1 : -1));
}

void WallAI::PassTime(double deltatime) {
	static const double AVERAGE_VELOCITY_COEF = 0.7;
	double vel = owner->maxvelocity * AVERAGE_VELOCITY_COEF;
	desiredposition += ForceVectorNorm(direction, vel) * deltatime;
}

void WallAI::Think(EntitySet &awareof) {
	double2 vect = owner->position / FIELD_SIZE - double2(0.5, 0.5);
	if ((vect ^ side) > 0.5) owner->hitpoints = -1.0;
	double2 diff = desiredposition - owner->position;
	owner->SetMoveDirection(diff);
	ShootGood(owner, FindEnemy(awareof));
}



DodgerAI::DodgerAI(Monster *_owner, double _awarenessradius
	, double _thinkperiod, double _playerimportance, double _randomimportance, double _missileimportance)
	 : AIBase(_owner, _awarenessradius)
	 , cooldown(_thinkperiod)
	 , playerimportance(_playerimportance), randomimportance(_randomimportance)
	 , missileimportance(_missileimportance) {
}
void DodgerAI::PassTime(double deltatime) {
	cooldown.PassTime(deltatime);
}

void DodgerAI::Think(EntitySet &awareof) {
	Player *pl = FindEnemy(awareof);
	if (cooldown.TimeToAct()) {
		double2 force;
		force += GetAttractionForce(owner, pl) * playerimportance;
		force += RandomVector() * randomimportance;
		force += GetAntiCollisionForce(owner, awareof);
		force += GetMissileDodgingForce(owner, awareof) * missileimportance;
		force += GetInFieldForce(owner) * 1.0;
		owner->SetMoveDirection(force);
	}
	ShootGood(owner, pl);
}
