#define _SECURE_SCL 0
#include "../cpp/world.h"
#include <cstdio>
#include <cstring>
#include <cassert>
#include <cmath>
#include <vector>
#undef min
#undef max
#include <algorithm>
using namespace std;

static const double EPS = 1e-9;
static const double INF = 1e+50;
static const double DELTATIME = 1.0 / 60.0;
static const double PI = 2.0 * acos(0.0);
static const double PIM2 = PI * 2.0;
static const double PID2 = PI / 2.0;

struct Point {
	double x, y;
	
	inline Point() : x(0), y(0) {}
	inline Point(double _x, double _y) : x(_x), y(_y) {}

	inline       double &operator[] (int i)       { return (i==0 ? x : y); }
	inline const double &operator[] (int i) const { return (i==0 ? x : y); }

	inline Point operator+ (const Point &b) const { return Point(x+b.x, y+b.y); }
	inline Point operator- (const Point &b) const { return Point(x-b.x, y-b.y); }
	inline Point operator* (double b) const { return Point(x*b, y*b); }
	inline Point operator/ (double b) const { return *this * (1.0/b); }

	inline void operator+= (const Point &b) { *this = *this + b; }
	inline void operator-= (const Point &b) { *this = *this - b; }
	inline void operator*= (double b) { *this = *this * b; }
	inline void operator/= (double b) { *this = *this / b; }

	inline double operator^ (const Point &b) const { return x*b.x + y*b.y; }
	inline double operator% (const Point &b) const { return x*b.y - y*b.x; }

	inline double operator!() const { return sqrt(*this ^ *this); }
};

inline double sqrd(const Point &x) { return x ^ x; }

Point Normal(const Point &x) {
	Point res = x;
	double norm = !res;
	if (norm > EPS) res /= norm;
	return res;
}

inline double trandom() { return rand() / 32768.0; }

Point RandomDirection() {
	Point res;
	do {
		res.x = 2.0 * trandom() - 1.0;
		res.y = 2.0 * trandom() - 1.0;
	} while (!res > 1.0 || !res < EPS);
	return Normal(res);
}

#define POINT(name) Point(name##X, name##Y)
#define COORDS(pnt) (pnt).x, (pnt.y)
static const int MONSTERS = World::MAXMONSTERS;
static const int MISSILES = World::MAXMISSILES;

extern World *world;


static const double MOVERADIUS = 0.4;
static const double SPREADTIME = 0.1;
static const double AWARERADIUS = 0.4;
static const double BITEWGT = 0.5;
static const double SHOOTANGLECOS = 0.70710678118654752440084436210485;
static const double PATHTIME = 5.0;

Point TargetMonsterDull(const CMonster &monster) {
	return POINT(monster.Position) - POINT(world->player.Position);
}

Point TargetMonsterMoving(const CMonster &monster) {
	Point pos = POINT(monster.Position) - POINT(world->player.Position);
	Point vel = POINT(monster.Velocity);
	pos += vel * (2.0*trandom()-1.0) * SPREADTIME;
	double speed = world->player.BulletVelocity;

	double qa = (vel ^ vel) - (speed * speed);
	double qb = (pos ^ vel);
	double qc = (pos ^ pos);
	if (qa > -EPS) return pos;
	double qd = qb*qb - qa*qc;
	qd = sqrt(qd);
	double t1 = (-qb - qd) / qa;
	double t2 = (-qb + qd) / qa;
	if (t1 < 0.0) t1 = INF;
	if (t2 < 0.0) t2 = INF;
	double t = std::min(t1, t2);

	return pos + vel*t;
}

Point ForceToField(Point pos, double rad) {
	if (pos.x < rad) pos.x = rad;
	if (pos.y < rad) pos.y = rad;
	if (pos.x > 1.0-rad) pos.x = 1.0-rad;
	if (pos.y > 1.0-rad) pos.y = 1.0-rad;
	return pos;
}

double MeleeRiskFunc(double dst, double mov, double bor) {
	if (dst <= bor + EPS) return 1.0;
	if (dst >= mov + bor - EPS) return -dst * EPS;
	return 1.0 - (dst - bor) / mov;
}
double GetRisk(Point pos, double dt) {
	double res = 0.0;
	for (int i = 0; i<MONSTERS; i++) {
		const CMonster &mon = world->monsters[i];
		if (mon.ClassName < 0) continue;

		Point delta = POINT(mon.Velocity)*0.5 * dt;
		Point mpos0 = POINT(mon.Position) - pos;
		Point mpos1 = mpos0 + delta;
		Point mpos2 = mpos1 + delta;

		double dst = sqrt(min(sqrd(mpos0), min(sqrd(mpos1), sqrd(mpos2))));
//		double dst = !mpos2;
		double mov = mon.MaxVelocity*0.5 * dt;
		double bor = mon.Radius + 2.0 * world->player.Radius + mon.MaxVelocity*0.3 * dt;

		double mult = EPS;
		mult = max(mult, 1.0*mon.BiteDamage);
		res += MeleeRiskFunc(dst, mov, bor) * mult;

		res += mon.BulletDamage * MeleeRiskFunc(dst, 1.0, bor);
	}
	for (int i = 0; i<MISSILES; i++) {
		const CMissile &miss = world->missiles[i];
		if (miss.ClassName < 0) continue;
		if (miss.ClassName == 200) continue;

		Point mpos = POINT(miss.Position) + POINT(miss.Velocity) * dt;
		double dst = !(mpos - pos);
		double bor = miss.Radius + 2.0 * world->player.Radius;

		double mult = EPS;
		if (miss.Damage > EPS) mult = miss.Damage * 4.0;
		res += MeleeRiskFunc(dst, world->player.Radius * 5.0, bor) * mult;
	}

	double bdst = min(min(pos.x, 1.0 - pos.x), min(pos.y, 1.0 - pos.y));
	bdst -= world->player.Radius;
	res += 0.1 * (0.5-bdst) * (0.5-bdst);

	double ctrd = !(pos * 2.0 - Point(1.0, 1.0));
	ctrd /= sqrt(2.0);
	if (ctrd > 0.95) res += 10.0 / (1.0 + EPS - ctrd);

	return res;
}

struct Path {
	int k;
	vector<Point> pos, vel;

	void Resize(int _k) {
		k = _k;
		pos.resize(k+1);
		vel.resize(k+1);
	}

	int Get(double t) {
		t *= k;
		int i = int(t);
		if (i < 0) i = 0;
		if (i > k-1) i = k-1;
		return i;
	}

	Point GetPos(double t) {
		int i = Get(t);
		return pos[i] * (1.0-t) + pos[i+1] * t;
	}
	Point GetVel(double t) {
		int i = Get(t);
		return vel[i];
	}
};

namespace ChooseBigPath {
	static const int PATHBIGPIECES = 6;
	static const int SCORESUBPIECES = 5;
	static const int PATHFIRSTANGLES = 10;
	static const int PATHANGLES = 3;

	static const double PATHBIGDT = PATHTIME / PATHBIGPIECES;

	Path best;
	double bestrisk;
	Path bigpath;
	double currentrisk;
	bool firstdescend;

	Path prev;
	
	void RecPath(int s) {
		if (currentrisk < bestrisk) {

			if (s == PATHBIGPIECES) {
				if (currentrisk < bestrisk) {
					bestrisk = currentrisk;
					best = bigpath;
				}
			}
			else {
				double offset = trandom() * PIM2;
				int cnt = (s==0 ? PATHFIRSTANGLES : PATHANGLES);
				for (int i = -firstdescend; i<cnt; i++) {
					if (i < 0) bigpath.vel[s+1] = prev.vel[s+1];
					else {
						double angle = offset + i * (PIM2 / cnt);
						bigpath.vel[s+1] = Point(cos(angle), sin(angle)) * world->player.MaxVelocity;
					}

					double wasrisk = currentrisk;
					Point p = bigpath.pos[s];
					for (int j = 0; j<SCORESUBPIECES; j++) {
						static const double SUBDT = PATHBIGDT/SCORESUBPIECES;
//						double t = (j + 0.5) / SCORESUBPIECES;
//						Point vel = bigpath.vel[s] * (1-t) + bigpath.vel[s+1] * t;
						Point vel = bigpath.vel[s+1];
						p = ForceToField(p + vel * SUBDT, world->player.Radius);
						currentrisk += GetRisk(p, s * PATHBIGDT + (j+1) * SUBDT);
					}
					bigpath.pos[s+1] = p;
					RecPath(s+1);
					currentrisk = wasrisk;
				}

/*
//				double offset = trandom() * PIM2;
				int cnt = (s==0 ? PATHFIRSTANGLES : PATHANGLES);
				for (int i = -firstdescend; i<cnt; i++) {
					if (i < 0) bigpath.vel[s+1] = prev.vel[s+1];
					else {
//						double angle = offset + i * (PIM2 / cnt);
						double rnd = trandom();
						double angle = rnd*rnd*rnd * PIM2;
						Point pvel = bigpath.vel[s];
						Point pnt(cos(angle), sin(angle));
						bigpath.vel[s+1] = Point(pnt % pvel, pnt ^ pvel);
					}

					double wasrisk = currentrisk;
					Point p = bigpath.pos[s];
					for (int j = 0; j<SCORESUBPIECES; j++) {
						static const double SUBDT = PATHBIGDT/SCORESUBPIECES;
						double t = (j + 0.5) / SCORESUBPIECES;
						Point vel = bigpath.vel[s] * (1-t) + bigpath.vel[s+1] * t;
						Point np = ForceToField(p + vel * SUBDT, world->player.Radius);
						currentrisk += GetRisk(p, s * PATHBIGDT + (j+1) * SUBDT);
						p = np;
					}
					bigpath.pos[s+1] = p;
					RecPath(s+1);
				}
                */
			}

		}
		firstdescend = false;
	}
	Path ChoosePath(Path _prev) {
		prev = _prev;
		if (prev.k == 0) prev.Resize(PATHBIGPIECES);
		firstdescend = true;

		Point startpos = POINT(world->player.Position);
		Point startvel = POINT(world->player.Velocity);
//		if (!startvel < EPS) startvel = RandomDirection();
//		startvel = Normal(startvel) * world->player.MaxVelocity;

		bigpath.Resize(PATHBIGPIECES);
		bigpath.pos[0] = startpos;
		bigpath.vel[0] = startvel;
		currentrisk = 0.0;
		bestrisk = INF;
		RecPath(0);
		assert(fabs(currentrisk) < EPS);
		return best;
	}
}

void BotThink() {
	//Moving
	static Path path;
	path = ChooseBigPath::ChoosePath(path);
	Point force = path.vel[1];

	force = Normal(force);
	world->run(COORDS(force));


	//Targeting
	double mind = 1e+50;
	Point best;
	for (int i = 0; i<MONSTERS; i++) if (world->monsters[i].ClassName >= 0) {
		Point tovect = TargetMonsterMoving(world->monsters[i]);
		double score = !tovect;

		double coef = 1.0;
		int cn = world->monsters[i].ClassName;
		if (cn == 105)	//dodger
			coef *= 5.0;
		if (cn == 104 || cn == 102)	//shooter
			coef *= 2.0;

		score /= coef;

		if ((force ^ Normal(tovect)) < SHOOTANGLECOS && world->monsters[i].BulletDamage < EPS) score *= 100.0;
		if (mind > score) {
			mind = score;
			best = tovect;
		}
	}
    
	world->shoot(COORDS(best));

}

#ifndef BOTWITHGLUT
World *world;
int main(int argc, char** argv) {
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 666;
	}
	world = new World(argv[1], atoi(argv[2]));
	world->makeMove();
	while (1) {
		BotThink();
		world->makeMove();
	}
	printf("Game over!\n");
	delete world;
	return 0;
}
#endif